@@ -41,6 +41,7 @@ extern int preserve_devices;
 extern int preserve_uid;
 extern int preserve_gid;
 extern int preserve_times;
+extern int relative_paths;
 
 static char **local_exclude_list = NULL;
 
@@ -158,8 +159,8 @@ void send_file_entry_v11(struct file_struct *file,int f)
   last_gid = file->gid;
   last_time = file->modtime;
 
-  strcpy(lastname,file->name);
-  lastname[255] = 0;
+  strncpy(lastname,file->name,MAXPATHLEN-1);
+  lastname[MAXPATHLEN-1] = 0;
 }
 
 
@@ -228,8 +229,8 @@ void receive_file_entry_v11(struct file_struct *file,
   last_gid = file->gid;
   last_time = file->modtime;
 
-  strcpy(lastname,file->name);
-  lastname[255] = 0;
+  strncpy(lastname,file->name,MAXPATHLEN-1);
+  lastname[MAXPATHLEN-1] = 0;
 }
 
 
@@ -315,15 +316,18 @@ static void send_file_name(int f,struct file_list *flist,char *fname)
   if (!file) return;  
   
   if (flist->count >= flist->malloced) {
-    flist->malloced += 100;
-    flist->files = (struct file_struct *)realloc(flist->files,
-						 sizeof(flist->files[0])*
-						 flist->malloced);
-    if (!flist->files)
-      out_of_memory("send_file_name");
+	  if (flist->malloced < 100)
+		  flist->malloced += 100;
+	  else
+		  flist->malloced *= 1.8;
+	  flist->files = (struct file_struct *)realloc(flist->files,
+						       sizeof(flist->files[0])*
+						       flist->malloced);
+	  if (!flist->files)
+		  out_of_memory("send_file_name");
   }
 
-  if (strcmp(file->name,".") && strcmp(file->name,"/")) {
+  if (strcmp(file->name,"/")) {
     flist->files[flist->count++] = *file;    
     send_file_entry(file,f);
   }
@@ -353,10 +357,13 @@ static void send_directory(int f,struct file_list *flist,char *dir)
     return;
   }
 
-  strcpy(fname,dir);
+  strncpy(fname,dir,MAXPATHLEN-1);
+  fname[MAXPATHLEN-1]=0;
   l = strlen(fname);
-  if (fname[l-1] != '/')
-    strcat(fname,"/");
+  if (fname[l-1] != '/') {
+	  strcat(fname,"/");
+	  l++;
+  }
   p = fname + strlen(fname);
 
   if (cvs_exclude) {
@@ -368,7 +375,7 @@ static void send_directory(int f,struct file_list *flist,char *dir)
     if (strcmp(di->d_name,".")==0 ||
 	strcmp(di->d_name,"..")==0)
       continue;
-    strcpy(p,di->d_name);
+    strncpy(p,di->d_name,MAXPATHLEN-(l+1));
     send_file_name(f,flist,fname);
   }
 
@@ -403,7 +410,8 @@ struct file_list *send_file_list(int f,int argc,char *argv[])
     char fname2[MAXPATHLEN];
     char *fname = fname2;
 
-    strcpy(fname,argv[i]);
+    strncpy(fname,argv[i],MAXPATHLEN-1);
+    fname[MAXPATHLEN-1] = 0;
 
     l = strlen(fname);
     if (l != 1 && fname[l-1] == '/') {
@@ -421,15 +429,19 @@ struct file_list *send_file_list(int f,int argc,char *argv[])
     }
 
     dir = NULL;
-    p = strrchr(fname,'/');
-    if (p) {
-      *p = 0;
-      if (p == fname) 
-	dir = "/";
-      else
-	dir = fname;      
-      fname = p+1;      
+
+    if (!relative_paths) {
+	    p = strrchr(fname,'/');
+	    if (p) {
+		    *p = 0;
+		    if (p == fname) 
+			    dir = "/";
+		    else
+			    dir = fname;      
+		    fname = p+1;      
+	    }
     }
+
     if (!*fname)
       fname = ".";
 
@@ -499,12 +511,15 @@ struct file_list *recv_file_list(int f)
     int i = flist->count;
 
     if (i >= flist->malloced) {
-      flist->malloced += 100;
-      flist->files =(struct file_struct *)realloc(flist->files,
-						  sizeof(flist->files[0])*
-						  flist->malloced);
-      if (!flist->files)
-	goto oom;
+	  if (flist->malloced < 100)
+		  flist->malloced += 100;
+	  else
+		  flist->malloced *= 1.8;
+	  flist->files =(struct file_struct *)realloc(flist->files,
+						      sizeof(flist->files[0])*
+						      flist->malloced);
+	  if (!flist->files)
+		  goto oom;
     }
 
     receive_file_entry(&flist->files[i],flags,f);
@@ -547,7 +562,7 @@ int file_compare(struct file_struct *f1,struct file_struct *f2)
 
 int flist_find(struct file_list *flist,struct file_struct *f)
 {
-  int low=0,high=flist->count;
+  int low=0,high=flist->count-1;
 
   while (low != high) {
     int mid = (low+high)/2;
