@@ -46,6 +46,7 @@ extern int recurse;
 extern int delete_mode;
 extern int cvs_exclude;
 extern int am_root;
+extern int relative_paths;
 
 /*
   free a sums struct
@@ -146,7 +147,6 @@ static struct sum_struct *receive_sums(int f)
   struct sum_struct *s;
   int i;
   off_t offset = 0;
-  int block_len;
 
   s = (struct sum_struct *)malloc(sizeof(*s));
   if (!s) out_of_memory("receive_sums");
@@ -160,8 +160,6 @@ static struct sum_struct *receive_sums(int f)
     fprintf(FERROR,"count=%d n=%d rem=%d\n",
 	    s->count,s->n,s->remainder);
 
-  block_len = s->n;
-
   if (s->count == 0) 
     return(s);
 
@@ -277,8 +275,14 @@ void recv_generator(char *fname,struct file_list *flist,int i,int f_out)
       }
       statret = -1;
     }
-    if (statret != 0 && mkdir(fname,file->mode) != 0 && errno != EEXIST)
-      fprintf(FERROR,"mkdir %s : %s\n",fname,strerror(errno));
+    if (statret != 0 && mkdir(fname,file->mode) != 0 && errno != EEXIST) {
+	    if (!(relative_paths && errno==ENOENT && 
+		  create_directory_path(fname)==0 && 
+		  mkdir(fname,file->mode)==0)) {
+		    fprintf(FERROR,"mkdir %s : %s (2)\n",
+			    fname,strerror(errno));
+	    }
+    }
     if (set_perms(fname,file,NULL,0) && verbose) 
       fprintf(FINFO,"%s/\n",fname);
     return;
@@ -523,23 +527,38 @@ static void delete_one(struct file_struct *f)
 }
 
 
+/* this deletes any files on the receiving side that are not present
+   on the sending side. For version 1.6.4 I have changed the behaviour
+   to match more closely what most people seem to expect of this option */
 static void delete_files(struct file_list *flist)
 {
   struct file_list *local_file_list;
   char *dot=".";
-  int i;
+  int i, j;
+  char *last_name=NULL;
 
   if (cvs_exclude)
     add_cvs_excludes();
 
-  if (!(local_file_list = send_file_list(-1,1,&dot)))
-    return;
+  for (j=0;j<flist->count;j++) {
+	  if (!S_ISDIR(flist->files[j].mode)) continue;
+	  if (strcmp(flist->files[j].name,".")==0) continue;
+	  if (last_name &&
+	      flist->files[j].name[strlen(last_name)] == '/' &&
+	      strncmp(flist->files[j].name,last_name, strlen(last_name))==0)
+		  continue;
+	  last_name = flist->files[j].name;
+	  if (verbose > 1)
+		  fprintf(FINFO,"deleting in %s\n", last_name);
+	  if (!(local_file_list = send_file_list(-1,1,&last_name)))
+		  return;
 
-  for (i=local_file_list->count-1;i>=0;i--) {
-    if (!local_file_list->files[i].name) continue;
-    if (-1 == flist_find(flist,&local_file_list->files[i])) {
-      delete_one(&local_file_list->files[i]);
-    }    
+	  for (i=local_file_list->count-1;i>=0;i--) {
+		  if (!local_file_list->files[i].name) continue;
+		  if (-1 == flist_find(flist,&local_file_list->files[i])) {
+			  delete_one(&local_file_list->files[i]);
+		  }    
+	  }
   }
 }
 
@@ -635,6 +654,10 @@ int recv_files(int f_in,struct file_list *flist,char *local_name,int f_gen)
       }
 
       /* open tmp file */
+      if (strlen(fname) > (MAXPATHLEN-8)) {
+	fprintf(FERROR,"filename too long\n");
+	continue;
+      }
       sprintf(fnametmp,"%s.XXXXXX",fname);
       if (NULL == mktemp(fnametmp)) {
 	fprintf(FERROR,"mktemp %s failed\n",fnametmp);
@@ -644,6 +667,10 @@ int recv_files(int f_in,struct file_list *flist,char *local_name,int f_gen)
 	continue;
       }
       fd2 = open(fnametmp,O_WRONLY|O_CREAT,file->mode);
+      if (relative_paths && errno == ENOENT && 
+	  create_directory_path(fnametmp) == 0) {
+	      fd2 = open(fnametmp,O_WRONLY|O_CREAT,file->mode);
+      }
       if (fd2 == -1) {
 	fprintf(FERROR,"open %s : %s\n",fnametmp,strerror(errno));
 	receive_data(f_in,buf,-1,NULL);
@@ -671,6 +698,10 @@ int recv_files(int f_in,struct file_list *flist,char *local_name,int f_gen)
 
       if (make_backups) {
 	char fnamebak[MAXPATHLEN];
+	if (strlen(fname) + strlen(backup_suffix) > (MAXPATHLEN-1)) {
+		fprintf(FERROR,"backup filename too long\n");
+		continue;
+	}
 	sprintf(fnamebak,"%s%s",fname,backup_suffix);
 	if (rename(fname,fnamebak) != 0 && errno != ENOENT) {
 	  fprintf(FERROR,"rename %s %s : %s\n",fname,fnamebak,strerror(errno));
@@ -698,16 +729,16 @@ int recv_files(int f_in,struct file_list *flist,char *local_name,int f_gen)
       }
     }
 
+  if (preserve_hard_links)
+	  do_hard_links(flist);
+
   /* now we need to fix any directory permissions that were 
      modified during the transfer */
-  if (!am_root) {
-    for (i = 0; i < flist->count; i++) {
-      struct file_struct *file = &flist->files[i];
-      if (!file->name || !S_ISDIR(file->mode)) continue;
-      recv_generator(file->name,flist,i,-1);
-    }
+  for (i = 0; i < flist->count; i++) {
+	  struct file_struct *file = &flist->files[i];
+	  if (!file->name || !S_ISDIR(file->mode)) continue;
+	  recv_generator(file->name,flist,i,-1);
   }
-  
 
   if (verbose > 2)
     fprintf(FERROR,"recv_files finished\n");
@@ -754,10 +785,11 @@ off_t send_files(struct file_list *flist,int f_out,int f_in)
 
       fname[0] = 0;
       if (file->dir) {
-	strcpy(fname,file->dir);
+	strncpy(fname,file->dir,MAXPATHLEN-1);
+	fname[MAXPATHLEN-1] = 0;
 	strcat(fname,"/");
       }
-      strcat(fname,file->name);
+      strncat(fname,file->name,MAXPATHLEN-strlen(fname));
 
       if (verbose > 2) 
 	fprintf(FERROR,"send_files(%d,%s)\n",i,fname);
